// (C) 2001-2020 Intel Corporation. All rights reserved.
// Your use of Intel Corporation's design tools, logic functions and other 
// software and tools, and its AMPP partner logic functions, and any output 
// files from any of the foregoing (including device programming or simulation 
// files), and any associated documentation or information are expressly subject 
// to the terms and conditions of the Intel Program License Subscription 
// Agreement, Intel FPGA IP License Agreement, or other applicable 
// license agreement, including, without limitation, that your use is for the 
// sole purpose of programming logic devices manufactured by Intel and sold by 
// Intel or its authorized distributors.  Please refer to the applicable 
// agreement for further details.


//-----------------------------------------------------------------------------
// This confidential and proprietary software may be used only as authorized by
// a licensing agreement from Intel
// copyright notice must be reproduced on all authorized copies.
//-----------------------------------------------------------------------------
// Copyright (C) 2017 Intel Corporation. All rights reserved.  Intel products are
// protected under numerous U.S. and foreign patents, maskwork rights, copyrights and
// other intellectual property laws.
//-----------------------------------------------------------------------------
//  Project Name:  whard_rock
//  Module Name :  wrpcie_tx_deskew
//  Author      :  tanjosh
//  Date        :  Feb 2018
//  Description :  Wrapper around deskew module for wrpcie cores
//                 - Includes glue logic for deskew interface
//                 - Parity generation logic for AIB signals from AIB adaptors
//                 to ip4/8/16 boundary
//-----------------------------------------------------------------------------

  module whr_rx_deskew #(
	parameter   DEVICE_WIDTH = "pcie_x16",
	parameter   PCIE_IP_MODE  = "pcie_x16",  // EPX16, EPX8, RPX4
    parameter   NUM_CHAN    = 6,       // # of AIB channels connected to Deskew module :PORT0:14, PORT1:7, PORT2:5, PORT3:5
    parameter   AIB_WIDTH   = 78,      // width of each AIB channel, including deskew marker bit
    parameter   DSK_WIDTH   = 77,      // ouptut datapath width for each AIB channel (if smaller than AIB_WIDTH, then datapath drops the MSBs)
    parameter   DSK_MARKER_POS  =  69, //bit position of deskew marker in incoming data
    parameter   DSK_DATA_PAR_WIDTH  =  64, //width of each AIB channel to be deskewed
    parameter   PARITY_SEG_WIDTH  =  8,    //Width of AIB channel bits used per parity bit
    parameter   DIS_MONITOR = 0        // 1 - removes the deskew Monitor feature
    //parameter   DIS_PARITY  = 0        // 1bit per NUM_CHAN - set '1' to disable parity generation on the particular channel
  )
  (input                                      pld_clk,
   input                                      pld_rst_n,

   //SW programmable CSRs for debug
   input                                      cfg_force_dsk_done_i,
   input                                      cfg_dsk_clr_i,

   input           [1:0]                      cfg_k_device_width_i,

   input           [NUM_CHAN*AIB_WIDTH-1:0]   aib_data_i,
   output  logic   [NUM_CHAN*DSK_WIDTH-1:0]   aib_data_deskewed_o,  //deskewed data with parity bits

   //CSR status
   output logic    [NUM_CHAN-1:0]             cfg_dsk_active_chans_o,
   output logic    [NUM_CHAN-1:0]             cfg_dsk_monitor_err_o,
   output logic                               cfg_dsk_mon_err_status_o,
   //FSR status signals
   output logic                               fsr_dsk_mon_err_status_o,
   output logic    [2:0]                      fsr_dsk_status_o,
   output logic                               fsr_dsk_vld_o,
   output logic                               fsr_dsk_eval_done_o
   );

   logic [NUM_CHAN-1:0] dsk_chan_sel;
   logic                dsk_clr_sync;
   logic                dsk_force_done_sync;
   logic [AIB_WIDTH*NUM_CHAN-1:0] aib_data_with_par;

   //To maintain consistent width of AIB_WIDTH across all channels,
   //extra bits from the incoming channel are passed to the deskew module
   //Parity is not generated for these extra aib signals
   //aib_data_with_parity[ch][AIB_WIDTH-1:0] = {deskew_marker, parity_bits, data_bits}
   localparam NUM_PARITY_BITS = DSK_DATA_PAR_WIDTH/PARITY_SEG_WIDTH;
   localparam EXTRA_AIB_DATA_WIDTH = (AIB_WIDTH-1-NUM_PARITY_BITS-DSK_DATA_PAR_WIDTH); //5
   localparam AIB_DATA_FOR_DSK_WIDTH = DSK_DATA_PAR_WIDTH + EXTRA_AIB_DATA_WIDTH; //69

   //Parity generation for incoming aib data
   //   -Append parity bits at MSB of every 64bit data group
   //Feedthrough deskew marker bit without parity generation
   genvar j, ch;
   generate
      for (ch=0; ch<NUM_CHAN; ch++) begin : loop_dsk_parity_gen
         for (j=0; j<NUM_PARITY_BITS; j++) begin : loop_dsk_parity_gen_inner
            //Generating 1bit parity per 8bits of aib data
            //i.e. total 8bits of parity for aib_data[63:0]
            assign aib_data_with_par[ch*AIB_WIDTH+DSK_DATA_PAR_WIDTH+EXTRA_AIB_DATA_WIDTH+j] = aib_data_i[ch*AIB_WIDTH+DSK_DATA_PAR_WIDTH+EXTRA_AIB_DATA_WIDTH+j+1];   // Parity generation not required on the rx side. Hence removed and forced all the parity bits to 0
                  // ^aib_data_i[(ch*AIB_WIDTH+j*PARITY_SEG_WIDTH)+:PARITY_SEG_WIDTH];
         end
         //Pass data to be deskewed (68:0)
         assign aib_data_with_par[ch*AIB_WIDTH+:AIB_DATA_FOR_DSK_WIDTH] =
                aib_data_i[ch*AIB_WIDTH+:AIB_DATA_FOR_DSK_WIDTH]; 
         //Pass incoming deskew marker (bit 69) to MSB (bit 77) of every channel
         assign aib_data_with_par[ch*AIB_WIDTH+DSK_WIDTH] = aib_data_i[ch*AIB_WIDTH+DSK_MARKER_POS];
      end
   endgenerate


   //Channel select logic
   localparam PORT0_1X16_CHAN_SEL = {15{1'b1}};
   localparam PORT0_2X8_CHAN_SEL  = {8{1'b1}};
   localparam PORT1_2X8_CHAN_SEL  = {8{1'b1}}; 
   localparam PORT0_4X4_CHAN_SEL  = {{5{1'b1}}};
   localparam PORT1_4X4_CHAN_SEL  = {{5{1'b1}}};
   localparam PORT2_4X4_CHAN_SEL  = {5{1'b1}}; 
   //localparam PORT3_4X4_CHAN_SEL  = {5{1'b1}}; 

   generate
      if (PCIE_IP_MODE == "pcie_x16")
         assign dsk_chan_sel = (DEVICE_WIDTH == "pcie_x16") ?  PORT0_1X16_CHAN_SEL :
                               (DEVICE_WIDTH == "pcie_x8x8") ?  PORT0_2X8_CHAN_SEL  :
							   (DEVICE_WIDTH == "pcie_x4x4x4x4") ?  PORT0_4X4_CHAN_SEL  :
                                '0;
      else if (PCIE_IP_MODE == "pcie_x8x8")
         assign dsk_chan_sel = (DEVICE_WIDTH == "pcie_x8x8") ?  PORT1_2X8_CHAN_SEL :
                               (DEVICE_WIDTH == "pcie_x4x4x4x4") ?  PORT1_4X4_CHAN_SEL :
                                '0;
      else if (PCIE_IP_MODE == "pcie_x4x4x4x4")
         assign dsk_chan_sel = (DEVICE_WIDTH == "pcie_x4x4x4x4") ?  PORT2_4X4_CHAN_SEL :
                               '0;
      else
         assign dsk_chan_sel = '0;
   endgenerate

   //Sync CSRs from cfg_clk to pld_clk
/*    wrpcie_sync3_bit #(.RESET_VAL(0),
                      .DST_CLK_FREQ_MHZ(1000), 
                      .SRC_DATA_FREQ_MHZ(500)) 
        sync_dsk_clr (.clk(pld_clk),
                      .rst_n(pld_rst_n),
                      .din_i(cfg_dsk_clr_i),
                      .dout_o(dsk_clr_sync));

   wrpcie_sync3_bit #(.RESET_VAL(0),
                      .DST_CLK_FREQ_MHZ(1000), 
                      .SRC_DATA_FREQ_MHZ(500)) 
        sync_dsk_force_done (.clk(pld_clk),
                             .rst_n(pld_rst_n),
                             .din_i(cfg_force_dsk_done_i),
                             .dout_o(dsk_force_done_sync)); */

   //Deskew module instance
   wrlib_rx_aib_deskew #(.NUM_CHAN(NUM_CHAN),
                         .AIB_WIDTH(AIB_WIDTH), //78,
                         .DSK_WIDTH(DSK_WIDTH), //77,
                         .DIS_MONITOR(DIS_MONITOR) //0
   ) u_wrpcie_deskew (.i_clk(pld_clk),
                      .i_rst_n(pld_rst_n),
                      .i_cfg_force_dsk_valid(cfg_force_dsk_done_i),
                      .i_cfg_dsk_chan_sel(dsk_chan_sel),
                      .i_cfg_wait_count_2x('0),
                      .i_dsk_clear(cfg_dsk_clr_i),
                      .i_aib_data(aib_data_with_par),
                      .o_aib_data_deskewed(aib_data_deskewed_o),
                      .o_dsk_active_chans(cfg_dsk_active_chans_o),
                      .o_dsk_monitor_err(cfg_dsk_monitor_err_o),
                      .o_dsk_monitor_err_status(fsr_dsk_mon_err_status_o),
                      .o_dsk_status(fsr_dsk_status_o),
                      .o_dsk_valid(fsr_dsk_vld_o),
                      .o_dsk_eval_done(fsr_dsk_eval_done_o));


   //dsk_err_status signal differentiates monitor_err & dsk_status  
   //reprsenting excessive_skew (0) and  monitor_err (1) 
   //cfg_dsk_mon_err_status_o : used in conjustion with cfg_dsk_monitor_err_o
   //fsr_dsk_mon_err_status_o : used in conjuction with fsr_dsk_status_o
   assign cfg_dsk_mon_err_status_o = fsr_dsk_mon_err_status_o;

endmodule
