// (C) 2001-2020 Intel Corporation. All rights reserved.
// Your use of Intel Corporation's design tools, logic functions and other 
// software and tools, and its AMPP partner logic functions, and any output 
// files from any of the foregoing (including device programming or simulation 
// files), and any associated documentation or information are expressly subject 
// to the terms and conditions of the Intel Program License Subscription 
// Agreement, Intel FPGA IP License Agreement, or other applicable 
// license agreement, including, without limitation, that your use is for the 
// sole purpose of programming logic devices manufactured by Intel and sold by 
// Intel or its authorized distributors.  Please refer to the applicable 
// agreement for further details.


//-----------------------------------------------------------------------------
// This confidential and proprietary software may be used only as authorized by
// a licensing agreement from ALTERA
// copyright notice must be reproduced on all authorized copies.
//-----------------------------------------------------------------------------
// Copyright Â© 2015 Altera Corporation. All rights reserved.  Altera products are
// protected under numerous U.S. and foreign patents, maskwork rights, copyrights and
// other intellectual property laws.
//-------------------------------------------------------------------------------
//  Project     :  
//  Module Name :                                   
//  Designed By : cmau/ngulston                                  
//  Date        :         
//  Description :  This module deskews the datapath across multiple AIB channels 
//                 Leveraged from Crete3 EHIP.  For WHR:
//                     - removed PTP/Elane support
//		       - removed references to EHIP
//                     - parameratized NUM_CHAN
//                     - Added parameter DIS_MONITOR
//                     - Added i_cfg_wait_count_2x
//                     - Added o_dsk_monitor_err_status
//                     - removed 'tx' from the top level ports
//
//  Usage notes:  
//       (1)  deskew markers should be  1 cycle wide, sent repetitively with 32-clk intervals
//
//       (2)  the SM will attempt deskew twice before allowing normal op (DESKEW_DONE), or giving up (DESKEW_ERROR state) 
//
//       (3)  success/failure is indicated via o_dsk_eval_done & o_dsk_status 
//
//       (4)  the user can restart the deskew SM at any time by asserting the synchronous clear pin "i_dsk_clear".
//               Recommended procedure if deskew keeps failing (-but should not be required):
//               (a) user stops sending deskew markers,
//               (b) user asserts i_dsk_clear, and holds it long enough for data pipes to flush
//               (c) user deasserts i_dsk_clear
//               (d) user starts sending deskew markers
//               
//       (5)  if (DIS_MONITOR=1), then after successful deskew, the SM stays in DESKEW_DONE, and ignores subsequent deskew markers
//            This is useful when the deskew marker bit is re-purposed after initial deskew.
//               
//       (6)  if (DIS_MONITOR=0), then after successful deskew, the SM continues to monitor the deskew markers and flags
//            an error if it detects a change-in-skew (MONITOR ERROR). 
//            When an error is detected, the SM will latch the post-deskew markers on o_dsk_monitor_err and go to the ERROR state
//            
//       (7)  The SM will go to the Error state if it fails to Deskew (including RETRY) -- this is 'EXCESSIVE SKEW', or 
//            if after Successful deskew, it detects a change in deskew (on post-deskew markers) -- this is 'MONITOR ERROR'.
//              (a) The SM stays in the Error state until cleared by i_dsk_clear, or on async reset
//              (b) While in the error state, the status of the channel markers is latched in o_dsk_monitor_err
//              (c) While in the error state, the o_dsk_status will indicate ERROR
//              (d) While in the error state, the datapath is by default disabled (o_dsk_valid=0), but
//                  can be forced to '1' via  i_cfg_force_dsk_valid=1 
//                  	                                                     
//-----------------------------------------------------------------------------
//Module Declaration
 
module wrlib_rx_aib_deskew #(
    parameter   NUM_CHAN  = 6,    // # of AIB channels
    parameter   AIB_WIDTH = 78,   // width of each AIB channel, including deskew marker bit
    parameter   DSK_WIDTH = 77,   // ouptut datapath width for each AIB channel (if smaller than AIB_WIDTH, then datapath drops the MSBs)
    parameter   DIS_MONITOR = 0   // 1 - removes the deskew Monitor feature  
    )

(
    //Clocks, Resets, and Config
    input   logic                              i_clk,
    input   logic                              i_rst_n, 
    input   logic                              i_cfg_force_dsk_valid,	   // For troubleshooting (default to 0) 
                                                                           // 1: When in ERROR state, force datapath to be enabled regardless of error
                                                                           // 0: When in ERROR state, disable datapath
    input   logic   [NUM_CHAN-1:0]             i_cfg_dsk_chan_sel,	   // bit significant.  '1' indicates which channel to include in deskew
    input   logic                              i_cfg_wait_count_2x,        // 1:  double the 'retry wait count' timer from 16 to 32 (for UPI)
                                                                           // 0:  use default 'retry wait count' timer value of 16

    //Deskew input
    input   logic                              i_dsk_clear,		   // global synchronous clear.   '1' clears the deskew machine and restarts the deskew process. 
    																			
    input   logic   [NUM_CHAN*AIB_WIDTH-1:0]   i_aib_data,		   // data received from an AIB channel, MSB must contain the deskew marker 
                            

    //Deskew output
    output  logic   [NUM_CHAN*DSK_WIDTH-1:0]   o_aib_data_deskewed,	  // deskewed datapath
    output  logic   [NUM_CHAN-1:0]             o_dsk_active_chans,	  // 1 bit per AIB chan.  real-time status - indicates which channels received a deskew marker 
                                                                          // Monitored at input of deskew machine         
									  
									  								    
									  
    output  logic  [NUM_CHAN-1:0]              o_dsk_monitor_err,         // 1 bit per AIB chan.  Value latches on error, and holds until SM is restarted via i_dsk_clear, or async reset
                                                                          // when o_dsk_status = ERROR & o_dsk_monitor_err_status=1,  
									  //      this vector indicates a MONITOR ERROR, and snapshots the markers that were NOT received 
									  //      on the output of the deskew machine when a MONITOR ERROR was detected. 
									  //      MONITOR ERROR=change in skew after the channels were successfully deskewed
									  //      this snapshot reflects post-deskew markers  
									  // when o_dsk_status = ERROR & o_dsk_monitor_err_status=0, this vector indicates EXCESSIVE SKEW ERROR,
									  //      and snapshots the channels that did NOT receive a marker when Excessive skew was detected.  
									  //      this snapshot reflects pre-deskew markers
									  
    output  logic                              o_dsk_monitor_err_status,  // when 1 & o_dsk_status indicates ERROR -- o_dsk_monitor_err/o_dsk_status corresponds to a Monitor Error
                                                                          // when 0 & o_dsk_status indicates ERROR -- o_dsk_monitor_err/o_dsk_status corresponds to Excessive Deskew
                                                                          // value is aligned with o_dsk_monitor_err, and holds until the SM is restarted via i_dsk_clear, or async reset    

    output  logic   [2:0]                      o_dsk_status, 	          // deskew status - decoded in conjunction with tx_dsk_eval_done
                                                        		  // 3'h7   = In DESKEW_ERROR state due to excessive skew, or monitor_err
					        			  // 3'h0-5 = amount of total chan-to-chan skew
									  // When deskew status = ERROR, remains in ERROR state until i_dsk_clear or async reset

    output  logic                              o_dsk_valid,  	          // 1 means data is successfully deskewed. level active.
                                                                          // may use this as a data-valid signal.
                                                                          // asserts 16 cycles after a successful deskew
									  // once asserted, stays asserted until Monitor error is detected, or when deskew is restarted via
									  // i_dsk_clear or async reset.
                                                                          // if (i_cfg_force_dsk_valid==1), then deasserts when a change-in-skew is detected 

    output  logic                              o_dsk_eval_done	          // 1 indicates deskew process complete.   0 indicates deskew process is active.  level active. success/failure is indicated by o_dsk_status 
                                                                          // once asserted, stays asserted until cleared via i_dsk_clear, or async reset
);

//-----------------------------------------------------------------------------
//Declarations
    logic     [NUM_CHAN-1:0]		       got_marker;
    logic     [NUM_CHAN-1:0][5:0]	       tap_sel_one_hot;
    logic     [NUM_CHAN-1:0]		       dsk_marker;
    logic    				       dsk_done;
    logic    				       reset_channels;
    logic     [NUM_CHAN-1:0][DSK_WIDTH-1:0]    dp_data;
    logic     [NUM_CHAN-1:0]		       dp_marker;
    logic     [NUM_CHAN-1:0][DSK_WIDTH-1:0]    dp_data_deskewed;    

    genvar    ww;

//-----------------------------------------------------------------------------
//Main Body of Code
 

    //=========================================================================
    //Create an instance of the deskew state machine for control/status

    wrlib_rx_aib_deskew_sm #(
        .CHANNELS    (NUM_CHAN   ),
        .DIS_MONITOR (DIS_MONITOR)
    )
    dsk_sm (

        //Clocks, Resets, and Config
        .i_clk                  (i_clk                  ),
        .i_rst_n                (i_rst_n                ),
        .i_dsk_clear            (i_dsk_clear            ),
        .i_cfg_force_dsk_valid  (i_cfg_force_dsk_valid   ),
	.i_cfg_wait_count_2x    (i_cfg_wait_count_2x    ),
        .i_use_channel          (i_cfg_dsk_chan_sel     ),  

        //Deskew_status from channels
        .i_got_marker           (got_marker             ),
        .i_tap_sel_one_hot      (tap_sel_one_hot        ),
        .i_dsk_marker           (dsk_marker             ),

        //Deskew control to channels
        .o_dsk_done             (dsk_done               ),
        .o_dsk_eval_done        (o_dsk_eval_done     ),
        .o_dsk_valid            (o_dsk_valid	     ),
        .o_dsk_active_chans     (o_dsk_active_chans  ),
        .o_reset_channels       (reset_channels         ),
        .o_dsk_status           (o_dsk_status	     ),
        .o_dsk_monitor_err      (o_dsk_monitor_err   ),
	.o_dsk_monitor_err_status (o_dsk_monitor_err_status)
    );


    //=========================================================================
    // Deskew Datapath

    generate
    for(ww=0;ww<NUM_CHAN;ww++)begin:dpchannels

        //Split the datapath back out into AIB channels. 
	// The deskew marker for each channel is the MSB of each channel
        assign  dp_data[ww]     =   i_aib_data[ww*AIB_WIDTH+:DSK_WIDTH];
        assign  dp_marker[ww]   =   i_aib_data[ww*AIB_WIDTH+AIB_WIDTH-1];

        wrlib_rx_aib_deskew_datapipe #(
            .WIDTH                  (DSK_WIDTH                      )
        )
        tx_aib_deskew_datapipe (
            .i_clk                  (i_clk                              ), 
            .i_rst_n                (i_rst_n                            ),  
            .i_dsk_clear            (reset_channels                     ),  
            .i_marker               (dp_marker[ww]                      ),
            .o_marker               (dsk_marker[ww]                     ),
            .i_dsk_done             (dsk_done                           ),  
            .i_aib_data             (dp_data[ww]                        ), 
            .o_aib_data_deskewed    (dp_data_deskewed[ww]               ),  
            .o_got_marker           (got_marker[ww]                     ), 
            .o_tap_sel_one_hot      (tap_sel_one_hot[ww]                ) 
        );

        assign  o_aib_data_deskewed[ww*DSK_WIDTH+:DSK_WIDTH] =   dp_data_deskewed[ww];

    end:dpchannels
    endgenerate


endmodule