// (C) 2001-2020 Intel Corporation. All rights reserved.
// Your use of Intel Corporation's design tools, logic functions and other 
// software and tools, and its AMPP partner logic functions, and any output 
// files from any of the foregoing (including device programming or simulation 
// files), and any associated documentation or information are expressly subject 
// to the terms and conditions of the Intel Program License Subscription 
// Agreement, Intel FPGA IP License Agreement, or other applicable 
// license agreement, including, without limitation, that your use is for the 
// sole purpose of programming logic devices manufactured by Intel and sold by 
// Intel or its authorized distributors.  Please refer to the applicable 
// agreement for further details.


//-----------------------------------------------------------------------------
// This confidential and proprietary software may be used only as authorized by
// a licensing agreement from ALTERA
// copyright notice must be reproduced on all authorized copies.
//-----------------------------------------------------------------------------
// Copyright Â© 2015 Altera Corporation. All rights reserved.  Altera products are
// protected under numerous U.S. and foreign patents, maskwork rights, copyrights and
// other intellectual property laws.
//-------------------------------------------------------------------------------
//  Project     :  
//  Module Name :                                   
//  Designed By :                                   
//  Date        :         
//  Description :  This module implements the delay pipe stages for a single AIB channel.
//                 The datapath is delayed in order to achieve channel-to-channel deskew.
//                 The # of delay stages necessary is determined by a dedicated Deskew Marker bit
//                 which comes in as part of the AIB data.                                                   
//-----------------------------------------------------------------------------------------------

module wrlib_rx_aib_deskew_datapipe 
#(
    parameter                   WIDTH   =   78           //Width of the data in the datapipe
)
(
    input   logic               i_clk,                  // System clock
    input   logic               i_rst_n,                // 0:async reset, must be deasserted synchronously
    input   logic               i_dsk_clear,            // restart the deskew process  
    input   logic               i_dsk_done,             // means all channels rcvd their deskew markers - freeze the tap selection    
    input   logic               i_marker,               // deskew marker input
    output  logic               o_marker,               // deskew marker output    
    input   logic   [WIDTH-1:0] i_aib_data,             // data received from an AIB channel 
    output  logic   [WIDTH-1:0] o_aib_data_deskewed,    // AIB data after deskew   
    output  logic               o_got_marker,           // means this channel received its deskew marker from the AIB input - once asserted, holds until deskew is restarted
                                                        // used by the top level controller to know when all AIB channels have received their markers
    output  logic   [5:0]       o_tap_sel_one_hot       // report number of delay stages inserted. one hot format.  bit0 = none, bit1=1 delay stage, bit2= 2 delay stages, bit3= 3 delays stages
                                                        // used by the top level controller to determine total skew 
);

//-----------------------------------------------------------------------------
//Declarations
     
    // encoding for tap_sel
    localparam SEL_INPUT   = 3'h0;   // no delay stage - pass the input directly to the output
    localparam SEL_DELAY_1 = 3'h1;   // tap the output from delay stage 1
    localparam SEL_DELAY_2 = 3'h2;   // tap the output from delay stage 2  
    localparam SEL_DELAY_3 = 3'h3;   // tap the output from delay stage 3
    localparam SEL_DELAY_4 = 3'h4;   // tap the output from delay stage 4
    localparam SEL_DELAY_5 = 3'h5;   // tap the output from delay stage 5
    
    //In the data pipe, the msb is the deskew marker bit
    localparam  DESKEW_MARKER_BIT   =   WIDTH;  
     
    logic   [4:0][WIDTH:0]      data_pipe; //  shift register for AIB data - used to insert skew on datapaths (1 deeper than expected max skew) 
    logic                       got_marker_r;   // indicates there is a deskew marker in the deskew_pipe.  
    logic   [2:0]               tap_sel;        // Chooses which delay stage to tap for the deskewed output data
    logic   [2:0]               tap_sel_r;


 logic [WIDTH-1:0] i_aib_data_r; 
 logic   [4:0][WIDTH:0]      data_pipe_r;
 logic i_marker_r;
 //logic o_got_marker_r;
//-----------------------------------------------------------------------------
//Main Body of Code
    
    always @ (posedge i_clk) begin
     data_pipe[1]        <=  i_dsk_clear?    '0:         data_pipe[0];   //stage 2
     data_pipe[2]        <=  i_dsk_clear?    '0:         data_pipe[1];   //stage 3
     data_pipe[3]        <=  i_dsk_clear?    '0:         data_pipe[2];   //stage 4
     data_pipe[4]        <=  i_dsk_clear?    '0:         data_pipe[3];   //stage 5
    end

    always @ (posedge i_clk)
    if (~i_rst_n) begin
        data_pipe[0]        <=  {(WIDTH+1){1'b0}}; 
     //   data_pipe[1]        <=  {(WIDTH+1){1'b0}}; 
      //  data_pipe[2]        <=  {(WIDTH+1){1'b0}}; 
      //  data_pipe[3]        <=  {(WIDTH+1){1'b0}}; 
      //  data_pipe[4]        <=  {(WIDTH+1){1'b0}}; 
        got_marker_r        <=  1'b0;
        tap_sel_r           <=  SEL_INPUT; 
        o_tap_sel_one_hot   <=  6'b00001;
        
        i_marker_r <= 1'b0;
        data_pipe_r[0] <= {(WIDTH+1){1'b0}}; 
        i_aib_data_r <= {(WIDTH){1'b0}};
        //o_got_marker <= 1'b0;
        
    end
    else begin 
        data_pipe[0]        <=  i_dsk_clear?    '0:         {i_marker,i_aib_data};     //stage 1
     //   data_pipe[1]        <=  i_dsk_clear?    '0:         data_pipe[0];   //stage 2
     //   data_pipe[2]        <=  i_dsk_clear?    '0:         data_pipe[1];   //stage 3
     //   data_pipe[3]        <=  i_dsk_clear?    '0:         data_pipe[2];   //stage 4
     //   data_pipe[4]        <=  i_dsk_clear?    '0:         data_pipe[3];   //stage 5


        got_marker_r        <=  i_dsk_clear?    '0:         o_got_marker;
        tap_sel_r           <=  i_dsk_clear?    SEL_INPUT:  tap_sel;
        

        //o_tap_sel_one_hot   <=  i_dsk_clear?                6'b000001:   // convert the tap_sel value into one-hot format
        //                        (tap_sel_r == SEL_INPUT)?     6'b000001:
        //                        (tap_sel_r == SEL_DELAY_1)?   6'b000010:
        //                        (tap_sel_r == SEL_DELAY_2)?   6'b000100:
        //                        (tap_sel_r == SEL_DELAY_3)?   6'b001000:
        //                        (tap_sel_r == SEL_DELAY_4)?   6'b010000:
        //                                                    6'b100000;
                                                            
        if(i_dsk_clear)
          o_tap_sel_one_hot <= 6'b000001;
        else begin
          o_tap_sel_one_hot <= 6'b100000;
          if(tap_sel_r == SEL_INPUT)
            o_tap_sel_one_hot <= 6'b000001;
          if(tap_sel_r == SEL_DELAY_1)
            o_tap_sel_one_hot <= 6'b000010;
          if(tap_sel_r == SEL_DELAY_2)
            o_tap_sel_one_hot <= 6'b000100;
          if(tap_sel_r == SEL_DELAY_3)
            o_tap_sel_one_hot <= 6'b001000;
          if(tap_sel_r == SEL_DELAY_4)
            o_tap_sel_one_hot <= 6'b010000;
        end

        i_marker_r <= i_marker;
        data_pipe_r <= data_pipe;
        i_aib_data_r <= i_aib_data;
        //o_got_marker <= o_got_marker_r;
                                                            
                                                            
    end
   

    // indicate that the deskew pulse was received by this channel
    // - set & hold when deskew pulse is received on input
    // - clear when deskew is restarted. 
    assign o_got_marker =   i_dsk_clear ?   1'b0 : 
                            i_marker ?      1'b1 : 
                                            got_marker_r;    

    // tap_sel specifies which pipe stage is used to drive the output datapath
    // - whichever stage contains the marker bit, is the stage to be used
    // - this value is latched when deskew is achieved
    // SEL_INPUT = use input stage
    // SEL_DELAY_1 = use 1st delay stage
    // SEL_DELAY_2 = use 2nd delay stage
    // SEL_DELAY_3 = use 3rd delay stage
    // SEL_DELAY_4 = use 4th delay stage
    // SEL_DELAY_5 = use 5th delay stage
    //assign tap_sel = i_dsk_clear                     ? SEL_INPUT:          // default to input stage when deskew is restarted.
    //                 i_dsk_done                      ? tap_sel_r :         // when all channels have a marker - freeze the tap_sel
    //                 i_marker   ?                      SEL_INPUT :         // actively deskewing, and marker bit is on the input
    //                 data_pipe[0][DESKEW_MARKER_BIT] ? SEL_DELAY_1:        // actively deskewing, and marker bit is in the first pipe stage 
    //                 data_pipe[1][DESKEW_MARKER_BIT] ? SEL_DELAY_2:        // actively deskewing, and marker bit is in the 2nd pipe stage
    //                 data_pipe[2][DESKEW_MARKER_BIT] ? SEL_DELAY_3:          
    //                 data_pipe[3][DESKEW_MARKER_BIT] ? SEL_DELAY_4:
    //                 data_pipe[4][DESKEW_MARKER_BIT] ? SEL_DELAY_5: 
    //                                                   tap_sel_r;  // if no marker in pipe, use last available information

    always @ (*) begin
      if(i_dsk_clear)
        tap_sel = SEL_INPUT;
      else begin 
        if(i_dsk_done)
          tap_sel = tap_sel_r;
        else begin
          tap_sel = tap_sel_r;
          if(i_marker)
            tap_sel = SEL_INPUT;
          if(data_pipe[0][DESKEW_MARKER_BIT])
            tap_sel = SEL_DELAY_1;
          if(data_pipe[1][DESKEW_MARKER_BIT])
            tap_sel = SEL_DELAY_2;
          if(data_pipe[2][DESKEW_MARKER_BIT])
            tap_sel = SEL_DELAY_3;
          if(data_pipe[3][DESKEW_MARKER_BIT])
            tap_sel = SEL_DELAY_4;
          if(data_pipe[4][DESKEW_MARKER_BIT])
            tap_sel = SEL_DELAY_5;
        end 
      end
    end

    // output data path - mux the output from one of the delay pipe/input stages
    //assign o_aib_data_deskewed = (tap_sel_r== SEL_DELAY_1) ? data_pipe_r[0][WIDTH-1:0] :
    //                             (tap_sel_r == SEL_DELAY_2) ? data_pipe_r[1][WIDTH-1:0] : 
    //                             (tap_sel_r == SEL_DELAY_3) ? data_pipe_r[2][WIDTH-1:0] : 
    //                             (tap_sel_r == SEL_DELAY_4) ? data_pipe_r[3][WIDTH-1:0] : 
    //                             (tap_sel_r == SEL_DELAY_5) ? data_pipe_r[4][WIDTH-1:0] :
    //                                                        i_aib_data_r;

    always @ (*) begin
      o_aib_data_deskewed = i_aib_data_r;
      if(tap_sel_r == SEL_DELAY_1)
        o_aib_data_deskewed = data_pipe_r[0][WIDTH-1:0];
      if(tap_sel_r == SEL_DELAY_2)
        o_aib_data_deskewed = data_pipe_r[1][WIDTH-1:0];
      if(tap_sel_r == SEL_DELAY_3)
        o_aib_data_deskewed = data_pipe_r[2][WIDTH-1:0];
      if(tap_sel_r == SEL_DELAY_4)
        o_aib_data_deskewed = data_pipe_r[3][WIDTH-1:0];
      if(tap_sel_r == SEL_DELAY_5)
        o_aib_data_deskewed = data_pipe_r[4][WIDTH-1:0];
    end

    always @ (*) begin
      o_marker = i_marker_r;
      if(tap_sel_r == SEL_DELAY_1)
        o_marker = data_pipe_r[0][WIDTH];
      if(tap_sel_r == SEL_DELAY_2)
        o_marker = data_pipe_r[1][WIDTH];
      if(tap_sel_r == SEL_DELAY_3)
        o_marker = data_pipe_r[2][WIDTH];
      if(tap_sel_r == SEL_DELAY_4)
        o_marker = data_pipe_r[3][WIDTH];
      if(tap_sel_r == SEL_DELAY_5)
        o_marker = data_pipe_r[4][WIDTH];
    end

    //assign o_marker             =   (tap_sel_r == SEL_DELAY_1) ? data_pipe_r[0][WIDTH] :
    //                                (tap_sel_r == SEL_DELAY_2) ? data_pipe_r[1][WIDTH] : 
    //                                (tap_sel_r == SEL_DELAY_3) ? data_pipe_r[2][WIDTH] : 
    //                                (tap_sel_r == SEL_DELAY_4) ? data_pipe_r[3][WIDTH] : 
    //                                (tap_sel_r == SEL_DELAY_5) ? data_pipe_r[4][WIDTH] :
    //                                                           i_marker_r;

endmodule
