// (C) 2001-2020 Intel Corporation. All rights reserved.
// Your use of Intel Corporation's design tools, logic functions and other 
// software and tools, and its AMPP partner logic functions, and any output 
// files from any of the foregoing (including device programming or simulation 
// files), and any associated documentation or information are expressly subject 
// to the terms and conditions of the Intel Program License Subscription 
// Agreement, Intel FPGA IP License Agreement, or other applicable 
// license agreement, including, without limitation, that your use is for the 
// sole purpose of programming logic devices manufactured by Intel and sold by 
// Intel or its authorized distributors.  Please refer to the applicable 
// agreement for further details.


//-----------------------------------------------------------------------------
// This confidential and proprietary software may be used only as authorized by
// a licensing agreement from ALTERA
// copyright notice must be reproduced on all authorized copies.
//-----------------------------------------------------------------------------
// Copyright Â© 2015 Altera Corporation. All rights reserved.  Altera products are
// protected under numerous U.S. and foreign patents, maskwork rights, copyrights and
// other intellectual property laws.
//-------------------------------------------------------------------------------
//  Project     :  
//  Module Name :                                   
//  Designed By : cmau/ngulston                                  
//  Date        :         
//  Description :  This module deskews the datapath across multiple AIB channels  
//                 Updates for WHR: 
//                     (1)  added parameter DIS_MONITOR  
//                     (2)  fixed i_cfg_force_dsk_valid to force dsk_valid=1 instead of resetting the datapipes
//                     (3)  added o_dsk_monitor_err_status to differentiate Excessive skew error 
//                          from Monitor error     
//                     (4)  Added  i_cfg_wait_count_2x  
//                     (5)  Added  debug_state   
//                     (6)  Fogbugz 567201 - Fixed bug in detecting excessive skew when markers are skewed 
//                          exactly 6 clks apart (rising edge to rising edge)       
//----------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------
//Module Declaration
 
module wrlib_rx_aib_deskew_sm 
#(
    parameter   CHANNELS    =   4,   //Number of AIB channels to deskew
    parameter   EHP_RX_DSK_STATUS_ERROR = 3'h7,
    parameter   DIS_MONITOR = 0     // 1 - disables deskew monitoring (after deskew done)
)
(
    //Clocks, Resets, and Config
    input   logic                                   i_clk,              //System clock
    input   logic                                   i_rst_n,            //0:async rst
    input   logic                                   i_dsk_clear,        //Synchronous clear for deskew logic
    input   logic                                   i_cfg_force_dsk_valid,     //Forces deskew to allow operation, even with error
    input   logic                                   i_cfg_wait_count_2x,      // 1:  set retry wait timer to 31, 0: set to default 15
    input   logic   [CHANNELS-1:0]                  i_use_channel,            //Indicates which channels must be deskewed

    //Deskew status from channels
    input   logic   [CHANNELS-1:0]                  i_got_marker,       //Sticky signal from deskew channels,1:deskew marker seen
    input   logic   [CHANNELS-1:0][5:0]             i_tap_sel_one_hot,  //Indicates which deskew stage each channel is in
    input   logic   [CHANNELS-1:0]                  i_dsk_marker,       //Non-sticky copy of the post-deskew deskew marker

    //Deskew control to channels
    output  logic                                   o_dsk_done,         //1:Freeze deskew channels
    output  logic                                   o_dsk_eval_done,    //1:Deskew evaluation done
    output  logic                                   o_dsk_valid,        //1:Deskew was good
    output  logic   [CHANNELS-1:0]                  o_dsk_active_chans, //Reflects got_marker
    output  logic                                   o_reset_channels,   //1:Sync reset of deskew channels
    output  logic   [2:0]                           o_dsk_status,       //Indicates max skew of channels after deskew;6 or 7 indicates error
    output  logic   [CHANNELS-1:0]                  o_dsk_monitor_err,  //If bonding is lost, indicates which channels failed
    output  logic                                   o_dsk_monitor_err_status // when 1 & o_dsk_status indicates ERROR -- o_dsk_monitor_err/o_dsk_status corresponds to a Monitor Error
                                                                             // when 0 & o_dsk_status indicates ERROR -- o_dsk_monitor_err/o_dsk_status corresponds to Excessive Deskew
     
);

//-----------------------------------------------------------------------------
//Declarations

    //State encodings
    localparam  ST_IDLE             =   3'd0;
    localparam  ST_BOND             =   3'd1;
    localparam  ST_FINISH_DESKEW    =   3'd2;
    localparam  ST_CLEAR_FOR_RETRY  =   3'd3;
    localparam  ST_RETRY            =   3'd4;
    localparam  ST_DESKEW_ERROR     =   3'd5;
    localparam  ST_WAIT_FOR_DONE    =   3'd6;
    localparam  ST_DESKEW_DONE      =   3'd7;

    logic                   dsk_done;
    logic   [2:0]           state;
    logic   [2:0]           nextstate;
    logic                   rx_eval_done;
    logic                   reset_retry;
    logic                   reset_channels;
    logic                   rx_dsk_valid;
    logic   [CHANNELS-1:0]  rx_dsk_monitor_err;
    logic   [4:0]           retry_wait_count;
    logic   [4:0]           retry_wait_limit;    
    logic   [5:0]           all_tap_sel_one_hot;
    logic                   all_got_marker;
    logic   [CHANNELS-1:0]  excessive_skew_lanes;
    logic                   got_excessive_skew;
    logic                   deskew_error;
    logic                   is_monitor_err;
    logic   [CHANNELS-1:0]  error_lanes;
    logic   [2:0]           rx_dsk_status;

    integer                 nn;

//-----------------------------------------------------------------------------
//Main Body of Code


    //=========================================================================
    //State machine logic 

    //State Register
    always_ff @(posedge i_clk)
    if(~i_rst_n)    state   <=  ST_IDLE;
    else            state   <=  i_dsk_clear?    ST_IDLE:    
                                                nextstate;

    //Nextstate and output logic
    always_comb begin:nextstate_logic
        case(state)
            
            //Wait for reset to be deasserted
            ST_IDLE:begin
                nextstate           =   ST_BOND;
                rx_eval_done        =   1'b0;
                reset_retry         =   1'b1;
                reset_channels      =   1'b1;
                rx_dsk_valid        =   1'b0;
                rx_dsk_monitor_err  =   '0;
		is_monitor_err      =   1'b0;
                dsk_done            =   1'b0;
            end

            //Wait for markers - first attempt to bond the channels
            ST_BOND:begin
                nextstate           =   all_got_marker?             ST_FINISH_DESKEW:
                                        got_excessive_skew?         ST_CLEAR_FOR_RETRY:
                                                                    ST_BOND;
                rx_eval_done        =   1'b0;
                reset_retry         =   1'b1;
                reset_channels      =   1'b0;
                rx_dsk_valid        =   1'b0;
                rx_dsk_monitor_err  =   '0;
		is_monitor_err      =   1'b0;
                dsk_done            =   all_got_marker?             1'b1:
                                                                    1'b0;
            end

            //Deskew markers detected, state machine freezes deskew channels
            //with required offsets and monitors for errors
            ST_FINISH_DESKEW:begin
                nextstate           =                               ST_WAIT_FOR_DONE;
                rx_eval_done        =   1'b0;
                reset_retry         =   1'b1;
                reset_channels      =   1'b0;
                rx_dsk_valid        =   1'b1;
                rx_dsk_monitor_err  =   '0;
		is_monitor_err  =   1'b0;
                dsk_done            =   1'b1;
            end
            
            //First deskew was unsuccessful, pause to flush out reset induced
            //skew and try again
            ST_CLEAR_FOR_RETRY:begin
                nextstate           =   (retry_wait_count< retry_wait_limit) ?  ST_CLEAR_FOR_RETRY:
                                                                    ST_RETRY;
                rx_eval_done        =   1'b0;
                reset_retry         =   1'b0;
                reset_channels      =   1'b1;
                rx_dsk_valid        =   1'b0;
                rx_dsk_monitor_err  =   '0;
		is_monitor_err      =   1'b0;
                dsk_done            =   1'b0;
            end

            //Wait for second chance deksew
            ST_RETRY:begin
                nextstate           =   all_got_marker?             ST_FINISH_DESKEW:
                                        got_excessive_skew?         ST_DESKEW_ERROR:
                                                                    ST_RETRY;
                rx_eval_done        =   1'b0;
                reset_retry         =   1'b1;
                reset_channels      =   1'b0;
                rx_dsk_valid        =   1'b0;  
                rx_dsk_monitor_err  =   got_excessive_skew?         excessive_skew_lanes:
                                                                    '0;   
		is_monitor_err      =   1'b0;             
                dsk_done            =   all_got_marker?             1'b1:
                                                                    1'b0;
            end

            //Deskew was unsuccessful, or deskew failed after successful first
            //deskew. Unless force deskew is high, hold the deskew lanes in
            //reset
            ST_DESKEW_ERROR:begin
                nextstate           =   ST_DESKEW_ERROR;
                rx_eval_done        =   1'b1;
                reset_retry         =   1'b1;
                reset_channels      =   1'b0;
                rx_dsk_valid        =   i_cfg_force_dsk_valid ? 1'b1 : 1'b0;
                rx_dsk_monitor_err  =   o_dsk_monitor_err; 
		is_monitor_err      =   o_dsk_monitor_err_status;
                dsk_done            =   1'b1;
            end

            //Wait until about 1/2 the deskew period before starting to check
            //for errors
            ST_WAIT_FOR_DONE:begin
                nextstate           =   (retry_wait_count < retry_wait_limit)? ST_WAIT_FOR_DONE:
                                                                    ST_DESKEW_DONE;
                rx_eval_done        =   1'b0;
                reset_retry         =   1'b0;
                reset_channels      =   1'b0;
                rx_dsk_valid        =   1'b1;
                rx_dsk_monitor_err  =   '0;
		is_monitor_err      =   1'b0;
                dsk_done            =   1'b1;
            end
            
            //Enter the deskew done state when deskew is complete. Stay here
            //unless an error is detected, in which case, record the lane
            //where the error occurred and go to the DESKEW_ERROR state
            ST_DESKEW_DONE:begin
                nextstate           =   DIS_MONITOR          ? ST_DESKEW_DONE : 
		                        deskew_error         ? ST_DESKEW_ERROR:
                                                               ST_DESKEW_DONE;
                rx_eval_done        =   1'b1;
                reset_retry         =   1'b1;
                reset_channels      =   1'b0;
                rx_dsk_valid        =   1'b1;
                rx_dsk_monitor_err  =   DIS_MONITOR  ? '0 : 
		                        deskew_error ?  error_lanes:
                                                       '0;
		is_monitor_err      =   DIS_MONITOR  ? '0 : 
		                        deskew_error ? '1 : '0;				       
                dsk_done            =   1'b1;
            end
            
            //Unknown input detected, output is unknown (x-pessimistic)
            default: begin    
                nextstate           =   'x; 
                rx_eval_done        =   'x;
                reset_retry         =   'x;
                reset_channels      =   'x;
                rx_dsk_valid        =   'x;
                rx_dsk_monitor_err  =   'x;
		is_monitor_err      =   'x;
                dsk_done            =   'x;
            end
        endcase
    end:nextstate_logic
                                
    
    //Register outputs
    always_ff @(posedge i_clk)
    if(~i_rst_n)begin
        o_dsk_eval_done      <=  '0;
        o_dsk_valid          <=  '0;
        o_dsk_active_chans   <=  '0;
        o_reset_channels     <=  1'b1;
        o_dsk_status         <=  '0;
        o_dsk_monitor_err    <=  '0;
	o_dsk_monitor_err_status <= 1'b0;
        o_dsk_done           <=  1'b0;
    end
    else begin
        o_dsk_eval_done      <=  rx_eval_done;
        o_dsk_valid          <=  rx_dsk_valid;
        o_dsk_active_chans   <=  (i_got_marker & i_use_channel);
        o_reset_channels     <=  reset_channels; 
        o_dsk_status         <=  rx_dsk_status;
        o_dsk_monitor_err    <=  rx_dsk_monitor_err;
	o_dsk_monitor_err_status <= is_monitor_err; 	
        o_dsk_done           <=  dsk_done;
    end


    //=========================================================================
    //Helper functions
    
    //Retry wait counter - counts 16 cycles to prevent retry from occurring
    //close to the next expected marker start position
    always_ff @(posedge i_clk)
    if(!i_rst_n)    retry_wait_count    <=  5'd0; 
    else            retry_wait_count    <=  reset_retry?    5'd0:
                                                            5'(retry_wait_count + 1'd1);
							    
    assign retry_wait_limit = i_cfg_wait_count_2x ? 5'd31 : 5'd15;
    
    //Combine and mask signals
    always_comb begin:combine_and_mask
        all_tap_sel_one_hot     =   6'd0;
        all_got_marker          =   1'b1;
        excessive_skew_lanes    =   '0;
        for(nn=0;nn<CHANNELS;nn++)begin:combine_loop
            all_tap_sel_one_hot         |=  ({6{i_use_channel[nn]}} & i_tap_sel_one_hot[nn]);
            all_got_marker              &=  (!i_use_channel[nn] | i_got_marker[nn]);
            excessive_skew_lanes[nn]    =   i_use_channel[nn] && !i_got_marker[nn];
        end:combine_loop
    end:combine_and_mask


    //Excessive skew occurs when at least one lane reaches its final pipe
    //stage before we have seen all markers
    // WHR:  'final pipe stage' is indicated when tap_sel_one_hot[4]=1.
    //        this is because the tap_sel_one_hot status is 1 cycle delayed
    //        from the internal datapipe value.
    assign  got_excessive_skew  =   all_tap_sel_one_hot[4] && !all_got_marker; 
     
    //A deskew error occurs when we get one expected marker without getting
    //all expected markers
    assign  deskew_error        =   |(i_dsk_marker & i_use_channel) //expected markers detected
                                    && ((i_dsk_marker & i_use_channel)!= i_use_channel); //...but not all


    //When a deskew error occurs, we mark all the lanes where a marker didn't
    //occur that when the other markers occurred as errors
    assign  error_lanes         =   i_use_channel & ~i_dsk_marker;


    //RX Dsk status indicates the result of the deskew
    assign  rx_dsk_status       =   (state == ST_DESKEW_ERROR)? EHP_RX_DSK_STATUS_ERROR:  // deskew error
                                    all_tap_sel_one_hot[5] ?    3'd5 :  // chan-to-chan skew is 5 
                                    all_tap_sel_one_hot[4] ?    3'd4 :  // chan-to-chan skew is 4 
                                    all_tap_sel_one_hot[3] ?    3'd3 :  // chan-to-chan skew is 3 
                                    all_tap_sel_one_hot[2] ?    3'd2 :  // chan-to-chan skew is 2
                                    all_tap_sel_one_hot[1] ?    3'd1 :  // chan-to-chan skew is 1
                                                                3'd0 ;  // else no skew

   logic[300:0] debug_state;
   always @ (state) begin
   case (state) 
       'd0: debug_state     = "ST_IDLE"; 
       'd1: debug_state     = "ST_BOND";
       'd2: debug_state     = "ST_FINISH_DESKEW";
       'd3: debug_state     = "ST_CLEAR_FOR_RETRY";
       'd4: debug_state     = "ST_RETRY";
       'd5: debug_state     = "ST_DESKEW_ERROR";
       'd6: debug_state     = "ST_WAIT_FOR_DONE";
       'd7: debug_state     = "ST_DESKEW_DONE";       
       default: debug_state = "UNKNOWN";
   
   endcase
   end

endmodule