// (C) 2001-2020 Intel Corporation. All rights reserved.
// Your use of Intel Corporation's design tools, logic functions and other 
// software and tools, and its AMPP partner logic functions, and any output 
// files from any of the foregoing (including device programming or simulation 
// files), and any associated documentation or information are expressly subject 
// to the terms and conditions of the Intel Program License Subscription 
// Agreement, Intel FPGA IP License Agreement, or other applicable 
// license agreement, including, without limitation, that your use is for the 
// sole purpose of programming logic devices manufactured by Intel and sold by 
// Intel or its authorized distributors.  Please refer to the applicable 
// agreement for further details.


/*
Description
- PCIECV CII workaround
- TD_1_7 AER Sev & Mask: use CII to modify CfgWr targeting AER_UNCOR_MSK[25:24] and AER_UNCOR_SEV[25:24] to be always 2'b00
- TD_1_43 ARI Control : use CII to force it 0
- Implements changes to fix ARI, AER, and SRIOV register issues.
*/

module ptile_pciecv_cii_workaround(
	//Clock
    input   wire                clk,
	//Reset
    input   wire                reset_n,
	//Error/corrupted
    input   wire                cii_hdr_poisoned,
	//Override
    output  wire                cii_override_en,
	//Function number
    input   wire    [2:0]       cii_func_num,
	//Virtualization - virtual function maps to one virtual machine
    input   wire    [10:0]      cii_vf_num,
	//Target physical or virtual
    input   wire                cii_wr_vf_active,
	//header first byte enable
    input   wire    [3:0]       cii_hdr_first_be,
	//Incoming Data
    input   wire    [31:0]      cii_dout,
	//Halt
    output  wire                cii_halt,
	//Request
    input   wire                cii_req,
	//Request Address
    input   wire    [9:0]       cii_addr,
	//write request
    input   wire                cii_write,
	//Output data
    output  wire    [31:0]      cii_override_din
);

localparam IPIPE = 4;
localparam OPIPE = 4;

//Address Offsets monitored by CII
localparam AER_UNCOR_MSK = 10'h42;
localparam AER_UNCOR_SEV = 10'h43;
localparam ARI_CONTROL   = 10'h5F;
localparam DEVICE_CTRL = 10'h26;
localparam LNK_STATUS = 10'h20;
localparam DEVICE_STATUS = 10'h1E;



//User Assigned Parameters
//Number of Functions
localparam MAX_F_NUM = 2;
//Enable SRIOV
localparam SRIOV_EN = 1'b1;
//ATS_ADDRESS based on topology: x16=C0 , x8=B2
localparam ATS = 10'hC0;

// input pipeline
reg     [IPIPE-1:0]         cii_hdr_poisoned_pipe;
// reg     [IPIPE*4-1:0]       cii_hdr_first_be_pipe;
reg     [IPIPE*32-1:0]      cii_dout_pipe;
reg     [IPIPE-1:0]         cii_req_pipe;
reg     [IPIPE*10-1:0]      cii_addr_pipe;
reg		[IPIPE-1:0]			cii_wr_vf_active_pipe;
reg     [IPIPE-1:0]         cii_write_pipe;
reg		[IPIPE*3-1:0]		cii_func_num_pipe;

wire                        cii_hdr_poisoned_int;
// wire    [3:0]               cii_hdr_first_be_int;
wire    [31:0]              cii_dout_int;
wire	[2:0]				cii_func_num_int;
wire                        cii_req_int;
wire    [9:0]               cii_addr_int;
wire							cii_wr_vf_active_int;
wire                        cii_write_int;

// edge detect and latch
reg                         cii_req_int_r;
reg                         cii_req_int_re;
reg                         cii_hdr_poisoned_int_r;
// reg     [3:0]               cii_hdr_first_be_int_r;
reg     [31:0]              cii_dout_int_r;
reg		[2:0]				cii_func_num_int_r;
reg     [9:0]               cii_addr_int_r;
reg							cii_wr_vf_active_int_r;
reg                         cii_write_int_r;

reg                         cii_req_int_lt;     // marker pulse
reg                         cii_hdr_poisoned_int_lt;
// reg     [3:0]               cii_hdr_first_be_int_lt;
reg     [31:0]              cii_dout_int_lt;
reg		[2:0]				cii_func_num_int_lt;
reg     [9:0]               cii_addr_int_lt;
reg   						cii_wr_vf_active_int_lt;
reg                         cii_write_int_lt;

// modify
reg                         valid_wr_req;
reg 						valid_rd_req;
reg                         aer_addr_hit;
reg                         ari_addr_hit;

//SRIOV registers
reg							link_sta_addr_hit;
reg							dev_ctrl_addr_hit;
reg							ats_addr_hit;
reg							dev_sta_addr_hit;			

reg                         cii_req_int_lt_r;     // marker pulse
reg     [31:0]              cii_dout_int_lt_r;
reg     [9:0]               cii_addr_int_lt_r;
reg							cii_wr_vf_active_int_lt_r;
reg		[2:0]				cii_func_num_int_lt_r;


reg                         cii_req_int_lt_r2;     // marker pulse
reg     [31:0]              cii_dout_int_lt_r2;
reg		[2:0]				cii_func_num_int_lt_r2;

reg                         cii_override_en_int;
reg                         cii_halt_int;
reg     [31:0]              cii_override_din_int;

// output pipeline
reg     [OPIPE-1:0]         cii_override_en_pipe;
reg     [OPIPE-1:0]         cii_halt_pipe;
reg     [OPIPE*32-1:0]      cii_override_din_pipe;

//==================================================================================

// input pipe with reset
always@( posedge clk or negedge reset_n )
begin
    if( ~reset_n )
    begin
        cii_req_pipe <= {IPIPE{1'b0}};
    end

    else
    begin
        cii_req_pipe <= {cii_req_pipe[0 +: IPIPE-1], cii_req};
    end
end

// input pipe
always@( posedge clk )
begin
    cii_hdr_poisoned_pipe <= {cii_hdr_poisoned_pipe[(IPIPE-1)*1-1:0], cii_hdr_poisoned};
    // cii_hdr_first_be_pipe <= {cii_hdr_first_be_pipe[(IPIPE-1)*4-1:0], cii_hdr_first_be};
    cii_dout_pipe         <= {cii_dout_pipe[(IPIPE-1)*32-1:0], cii_dout};
    cii_addr_pipe         <= {cii_addr_pipe[(IPIPE-1)*10-1:0], cii_addr};
	cii_wr_vf_active_pipe <= {cii_wr_vf_active_pipe[(IPIPE-1)-1:0], cii_wr_vf_active}; 
    cii_write_pipe        <= {cii_write_pipe[(IPIPE-1)*1-1:0], cii_write};
	cii_func_num_pipe	  <= {cii_func_num_pipe[(IPIPE-1)*3-1:0], cii_func_num};
end

assign cii_hdr_poisoned_int = cii_hdr_poisoned_pipe[(IPIPE-1)*1 +: 1];
// assign cii_hdr_first_be_int = cii_hdr_first_be_pipe[(IPIPE-1)*4 +: 4];
assign cii_dout_int         = cii_dout_pipe[(IPIPE-1)*32 +: 32];
assign cii_func_num_int		= cii_func_num_pipe[(IPIPE-1)*3 +: 3];
assign cii_req_int          = cii_req_pipe[(IPIPE-1)*1 +: 1];
assign cii_addr_int         = cii_addr_pipe[(IPIPE-1)*10 +: 10];
assign cii_wr_vf_active_int = cii_wr_vf_active_pipe[(IPIPE-1) +: 1];
assign cii_write_int        = cii_write_pipe[(IPIPE-1)*1 +: 1];

// edge detect and latch
always@( posedge clk or negedge reset_n )
begin
    if( ~reset_n )
    begin
        cii_req_int_r  <= 1'b0;
        cii_req_int_re <= 1'b0;
        cii_req_int_lt <= 1'b0;
    end
    else
    begin
        cii_req_int_r  <= cii_req_int;
        cii_req_int_re <= ( ~cii_req_int_r & cii_req_int ) ? 1'b1 : 1'b0;
        cii_req_int_lt <= cii_req_int_re;
    end
end

always@( posedge clk )
begin
    cii_hdr_poisoned_int_r <= cii_hdr_poisoned_int;
    // cii_hdr_first_be_int_r <= cii_hdr_first_be_int;
    cii_dout_int_r         <= cii_dout_int;
    cii_addr_int_r         <= cii_addr_int;
	cii_wr_vf_active_int_r <= cii_wr_vf_active_int;
    cii_write_int_r        <= cii_write_int;
	cii_func_num_int_r		<= cii_func_num_int;

    if( cii_req_int_re )
    begin
        cii_hdr_poisoned_int_lt <= cii_hdr_poisoned_int_r;
        // cii_hdr_first_be_int_lt <= cii_hdr_first_be_int_r;
        cii_dout_int_lt         <= cii_dout_int_r;
        cii_addr_int_lt         <= cii_addr_int_r;
		cii_wr_vf_active_int_lt <= cii_wr_vf_active_int_r;
        cii_write_int_lt        <= cii_write_int_r;
		cii_func_num_int_lt			<= cii_func_num_int_r;
    end
end

// modify
// reg
always@( posedge clk or negedge reset_n )
begin
    if( ~reset_n )
    begin
        cii_req_int_lt_r  <= 1'b0;
        cii_req_int_lt_r2 <= 1'b0;
    end
    else
    begin
        cii_req_int_lt_r  <= cii_req_int_lt;
        cii_req_int_lt_r2 <= cii_req_int_lt_r;
    end
end

always@( posedge clk )
begin
    cii_dout_int_lt_r         <= cii_dout_int_lt;
    cii_dout_int_lt_r2         <= cii_dout_int_lt_r;
	
	cii_func_num_int_lt_r		<= cii_func_num_int_lt;
	cii_func_num_int_lt_r2		<= cii_func_num_int_lt_r;
	
    cii_addr_int_lt_r         <= cii_addr_int_lt;
	cii_wr_vf_active_int_lt_r		  <= cii_wr_vf_active_int_lt;
end

// stage 1 - valid read and write requests
always@( posedge clk or negedge reset_n )
begin
    if( ~reset_n )
    begin
        valid_wr_req <= 1'b0;
		valid_rd_req <= 1'b0;
    end

    else
    begin
        valid_wr_req <= ( cii_req_int_lt & cii_write_int_lt & ~cii_hdr_poisoned_int_lt ) ? 1'b1 : 1'b0;
		valid_rd_req <= ( cii_req_int_lt & ~cii_write_int_lt & ~cii_hdr_poisoned_int_lt ) ? 1'b1 : 1'b0;
    end
end

// stage 2 - check for monitored address hit
always@( posedge clk or negedge reset_n )
begin
    if( ~reset_n )
    begin
        aer_addr_hit <= 1'b0;
        ari_addr_hit <= 1'b0;
		link_sta_addr_hit <= 1'b0;
		dev_ctrl_addr_hit <= 1'b0;
		ats_addr_hit <= 1'b0;
		dev_sta_addr_hit <= 1'b0;
    end

    else
    begin
        if ( valid_wr_req )
        begin
            aer_addr_hit <= ( cii_addr_int_lt_r == AER_UNCOR_MSK || cii_addr_int_lt_r == AER_UNCOR_SEV ) ? 1'b1 : 1'b0;
            ari_addr_hit <= ( cii_addr_int_lt_r == ARI_CONTROL && cii_wr_vf_active_int_lt_r == 1'b0 ) ? 1'b1 : 1'b0;
			dev_sta_addr_hit <= (cii_addr_int_lt_r == DEVICE_STATUS && cii_wr_vf_active_int_lt_r) ? 1'b1 : 1'b0;
        end
        else if (valid_rd_req)
        begin
            aer_addr_hit <= 1'b0;
            ari_addr_hit <= ( cii_addr_int_lt_r == ARI_CONTROL ) ? 1'b1 : 1'b0;
			link_sta_addr_hit <= (cii_addr_int_lt_r == LNK_STATUS && SRIOV_EN && cii_wr_vf_active_int_lt_r) ? 1'b1 : 1'b0;
			dev_ctrl_addr_hit <= (cii_addr_int_lt_r == DEVICE_CTRL && SRIOV_EN && cii_wr_vf_active_int_lt_r) ? 1'b1 : 1'b0;
			ats_addr_hit <= (cii_addr_int_lt_r == ATS && SRIOV_EN && cii_wr_vf_active_int_lt_r) ? 1'b1 : 1'b0;
        end
		else
		begin
			aer_addr_hit <= 1'b0;
            ari_addr_hit <= 1'b0;
			link_sta_addr_hit <= 1'b0;
			dev_ctrl_addr_hit <= 1'b0;
			ats_addr_hit <= 1'b0;
			dev_sta_addr_hit <= 1'b0;
		end
    end
end

// stage 3 - mask write to bit[25:24]
always@( posedge clk or negedge reset_n )
begin
    if( ~reset_n )
    begin
        cii_halt_int <= 1'b1;
        cii_override_en_int <= 1'b0;
    end

    else
    begin
        cii_halt_int <= ( cii_req_int_lt_r2 ) ? 1'b0 : 1'b1;
        cii_override_en_int <= ( aer_addr_hit | ari_addr_hit | link_sta_addr_hit |	
		dev_ctrl_addr_hit |	ats_addr_hit | dev_sta_addr_hit) ? 1'b1 : 1'b0;
    end
end

always@( posedge clk )
begin
    if( aer_addr_hit )
        cii_override_din_int <= cii_dout_int_lt_r2 & 32'hfcff_ffff;
    //FIX: ARI is hit, send out next function number or 0
	else if( ari_addr_hit )
	begin
		cii_override_din_int <=	
		( MAX_F_NUM - 1 > cii_func_num ) ? (cii_dout_int_lt_r2 & 32'hffff_00ff) | (cii_func_num_int_lt_r2 + 1)<<8: (cii_dout_int_lt_r2 & 32'hffff_00ff) ; 
	end
	else if( link_sta_addr_hit )
	begin
		// [16:31] = 0
		cii_override_din_int <= cii_dout_int_lt_r2 & 32'h0000_ffff;
	end
	else if( dev_ctrl_addr_hit )
	begin
		//[12] = 0
		cii_override_din_int <= cii_dout_int_lt_r2 & 32'hffff_efff; 
	end
	else if( ats_addr_hit )
	begin
		// [16:20] = 0
		cii_override_din_int <= cii_dout_int_lt_r2 & 32'hffe0_ffff;
	end
	else if( dev_sta_addr_hit )
	begin
		//[20] = 0
		cii_override_din_int <= cii_dout_int_lt_r2 & 32'hffef_ffff;
	end
	else
        cii_override_din_int <= cii_dout_int_lt_r2 & 32'h0000_ffff;
end

// output pipe with reset
always@( posedge clk or negedge reset_n )
begin
    if( ~reset_n )
    begin
        cii_override_en_pipe <= {OPIPE{1'b0}};
        cii_halt_pipe <= {OPIPE{1'b1}};
    end

    else
    begin
        cii_override_en_pipe <= {cii_override_en_pipe[(OPIPE-1)*1-1:0], cii_override_en_int};
        cii_halt_pipe <= {cii_halt_pipe[(OPIPE-1)*1-1:0], cii_halt_int};
    end
end

// output pipe
always@( posedge clk )
begin
    cii_override_din_pipe <= {cii_override_din_pipe[(OPIPE-1)*32-1:0], cii_override_din_int};
end

assign cii_override_en  = cii_override_en_pipe[(OPIPE-1)*1 +: 1];
assign cii_halt         = cii_halt_pipe[(OPIPE-1)*1 +: 1];
assign cii_override_din = cii_override_din_pipe[(OPIPE-1)*32 +: 32];

endmodule
