# (C) 2001-2020 Intel Corporation. All rights reserved.
# Your use of Intel Corporation's design tools, logic functions and other 
# software and tools, and its AMPP partner logic functions, and any output 
# files from any of the foregoing (including device programming or simulation 
# files), and any associated documentation or information are expressly subject 
# to the terms and conditions of the Intel Program License Subscription 
# Agreement, Intel FPGA IP License Agreement, or other applicable 
# license agreement, including, without limitation, that your use is for the 
# sole purpose of programming logic devices manufactured by Intel and sold by 
# Intel or its authorized distributors.  Please refer to the applicable 
# agreement for further details.



# -----------------------------------------------------------------------------
# This procedure constrains the max_delay (not skew) between the status bit regs.
#
# The hierarchy path to the status_bits CDC instance is required as an 
# argument.
# -----------------------------------------------------------------------------
proc constrain_data_bits_max_delay { path } {

    #set the to/from paths for stp_status_bits
    set path_from $path|stp_status_bits_in_reg_acq\[*\]
    set path_to $path|stp_status_bits_out_reg_tck\[*\]

    #check if the paths to be constrained exist or not
    set num_status_paths_from [get_collection_size [get_registers -nowarn $path_from]]
    set num_status_paths_to [get_collection_size [get_registers -nowarn $path_to]]

    #if either "to" or "from" paths donot exist, exit the .sdc gracefully
    if {$num_status_paths_from > 0 && $num_status_paths_to > 0} {
        #call to function to get the tck domain name and period
        ##puts "DEBUG: my path = $path|stp_status_bits_out_reg_tck*"
        set max_delay_prd [expr [get_clk_info $path|stp_status_bits_out_reg_tck*]]
        ##puts "DEBUG: max delay is 1xclk_prd = $max_delay_prd"

        #set the max delay as function of dst clk period
        #max delay is 1xdsk clk period (because valid bit takes ~3 cycles in dst clk to go from src clk domain)
        set_max_delay -from [ get_registers $path_from ] -to [ get_registers $path_to ] $max_delay_prd
        set_max_skew -to [ get_registers $path_from ] $max_delay_prd
    }  

}

# -----------------------------------------------------------------------------
# This procedure is to find out the tck clk name and period
#
# The hierarchy path to the data_bits CDC instance is required as an 
# argument.
# -----------------------------------------------------------------------------
proc get_clk_info { filter } {
    ##puts "DEBUG: Search for $filter"
    set my_clk [get_fanins $filter -clock -stop_at_clocks]
    ##puts "DEBUG: my_clk = $my_clk"

    foreach_in_collection clk $my_clk {
        set clk_node_name [get_node_info -name $clk]
        set clks [get_clocks_feeding_register $filter]
        if {[llength $clks] == 0} {
            ##puts "DEBUG: clk clk period is not defined, setting max delay to 5ns"
            post_message -type warning "In-System Source and Probe synchronous transfer clock is provided in '[get_current_instance]', but its clock period is not defined on '$clk_node_name'. If the clock period is defined in the project, please change the SDC file order such that the clock definition is evaluated before the In-System Source and Probe .sdc/.ip file. As a fallback, the delay is set based on the 5ns clock period (default 200MHz clock)"
            set clk_prd 5
            ##puts "DEBUG: tck domain period (default): $clk_prd"
        } else {
            # In the case of multiple clock definitions, arbitrarily use the first clock in the list
            set clk [lindex $clks 0]
            set clk_prd [get_clock_info [get_clocks $clk] -period]
            post_message -type info "In-System Source and Probe synchronous transfer clock is provided in '[get_current_instance]'. The clock period $clk_prd (ns) is defined on '$clk_node_name'. Data delay constraints are set on the CDC data path to/from the source/probe ports accordingly."
            ##puts "DEBUG: tck domain period: $clk_prd"  
        }
    }

    return $clk_prd
}

#
# Implementation is referenced from the script in the case 1507838137
#
proc get_clocks_feeding_register { reg_name } {
    set ret [list]
    
    # Pick the first register in the matching list.
    set reg_col [get_registers $reg_name]
    foreach_in_collection reg $reg_col {
        set reg_name [get_object_info -name $reg]
        break
    }
    
    catch { array unset nodes_with_clocks }
    catch { array unset node_types }
    array set nodes_with_clocks [list]
    array set node_types [list]
    set register_drivers [list]
    
    # We may want to return the clock objects for further processing
    catch { array unset clock_to_object }
    array set clock_to_object [list]
  
    # Set up a mapping for any node in the design with a clock on it
    # Map from the node name to a list of clocks on the node
    foreach_in_collection clock_id [all_clocks] {
        set clock_name [get_clock_info -name $clock_id]
        set clock_target_col [get_clock_info -targets $clock_id]
        set clock_to_object($clock_name) $clock_id
    
        # Skip any virtual clocks - they never have targets
        if { 0 < [get_collection_size $clock_target_col] } {
            foreach_in_collection target_id [get_clock_info -targets $clock_id] {
                set target_name [get_node_info -name $target_id]
                set target_type [get_node_info -type $target_id]
                lappend nodes_with_clocks($target_name) $clock_name
                set node_types($target_name) $target_type
            }
        }
    }

    # At least one node with a clock on it is in the register's clock fanin path.
    # We can be clever with the use of get_fanins to check for the number
    # of fanin nodes through a node with a clock on it to the clock
    # port of the desired register. If there is at least one fanin, that node
    # is in the register's clock fanin path.
    foreach node_with_a_clock [array names nodes_with_clocks] {

        # I have to explicitly include collection calls, because -through
        # says there are no fanins if I don't make it a collection.
        switch -exact -- $node_types($node_with_a_clock) {
            "pin" {  set through_col [get_pins $node_with_a_clock] }
            "port" {
                set fanout_edges [get_node_info -fanout_edges [foreach_in_collection port [get_ports $node_with_a_clock] { break }; expr {$port}]]
                if {$fanout_edges == ""} { continue }
                set pin_filter []
                foreach edge $fanout_edges {
                    lappend pin_filter [get_object_info -name [get_edge_info -dst $edge]]
                }
                set through_col [get_pins -nowarn $pin_filter]
                if { 0 == [get_collection_size $through_col] } { continue }
            }
            "cell" { set through_col [get_cells $node_with_a_clock] }
            default { continue }
        }
        set fanin_col [get_fanins -through $through_col ${reg_name}|clk]
        if { 0 < [get_collection_size $fanin_col] } {
            # Good, node_with_a_clock is in the register's clock fanin path
            lappend register_drivers $node_with_a_clock
        } else {
            # No nodes fan in to the register through node_with_a_clock, so the
            # node doesn't clock the register.
        }

    }

    # Make sure we actually have at least one node driving the register
    # before we continue
    if { [llength $register_drivers] > 0 } {
        # We want to know the last node that drives the register.
        # If there's only one in the list of register drivers, then we're done.
        # There can be multiple nodes that drive the register, because
        # a top-level IO that feeds a PLL that feeds the register
        # is considered to fanin to the register, and the same with the PLL.
        while { 1 < [llength $register_drivers] } {

            set node_a [lindex $register_drivers 0]
            set node_b [lindex $register_drivers 1]
            
            # I have to explicitly include collection calls, because -through
            # says there are no fanins if I don't make it a collection.
            switch -exact -- $node_types($node_b) {
                "pin" {  set through_col [get_pins $node_b] }
                "port" { set through_col [get_ports $node_b] }
                "cell" { set through_col [get_cells $node_b] }
                "reg" {  set through_col [get_registers $node_b] }
                default { continue }
            }
            
            # If node_b is on the fanin path to node_a, then node_a
            # is closer to the register than node_b in the clock tree.
            # If there's no fanin through node_b through node_a,
            # then node_a comes before node_b in the clock tree,
            set fanin_col [get_fanins -clock -through $through_col $node_a]
            
            if { 0 < [get_collection_size $fanin_col] } {
                # node_a is closer to the register.
                # get rid of node_b
                set register_drivers [lreplace $register_drivers 1 1]
            } else {
                # node_b is closer to the register
                # get rid of node_a
                set register_drivers [lrange $register_drivers 1 end]
            }
        }
        set node_driving_register [lindex $register_drivers 0]
        set ret $nodes_with_clocks($node_driving_register)
    }
    
    return $ret
}

constrain_data_bits_max_delay "[get_current_instance]|issp_impl|altsource_probe_body_inst|*|cdc_gen.probe_cdc_gen.probe_cdc"
constrain_data_bits_max_delay "[get_current_instance]|issp_impl|altsource_probe_body_inst|*|cdc_gen.src_cdc_gen.src_cdc"
